import json
import logging
import pathlib
from typing import Any, Dict, List, Optional, Union

import pyhf

from . import configuration
from . import histo


log = logging.getLogger(__name__)


class WorkspaceBuilder:
    """Collects functionality to build a ``pyhf`` workspace."""

    def __init__(self, config: Dict[str, Any]) -> None:
        """Creates a workspace corresponding to a cabinetry configuration.

        Args:
            config (Dict[str, Any]): ``cabinetry`` configuration
        """
        self.config = config
        self.histogram_folder = pathlib.Path(config["General"]["HistogramFolder"])

    def _get_data_sample(self) -> Dict[str, Any]:
        """Returns the data sample dictionary.

        Returns:
            Dict[str, Any]: the data sample dictionary
        """
        data_samples = [
            sample for sample in self.config["Samples"] if sample.get("Data", False)
        ]
        if len(data_samples) != 1:
            raise ValueError("did not find exactly one data sample")
        return data_samples[0]

    def _get_constant_parameter_setting(self, par_name: str) -> Optional[float]:
        """Determines whether a parameter should be set to constant, and to which value.

        For a given parameter, determines if it is supposed to be set to constant. If
        not, returns None, otherwise returns the value it should be fixed to. This only
        looks for the first occurrence of the parameter in the list.

        Args:
            par_name (str): name of parameter to check

        Returns:
            Optional[float]: returns None if parameter is not supposed to be held
            constant, or the value it has to be fixed to
        """
        fixed_parameters = self.config["General"].get("Fixed", [])
        fixed_value = next(
            (
                fixed_par["Value"]
                for fixed_par in fixed_parameters
                if fixed_par["Name"] == par_name
            ),
            None,
        )
        return fixed_value

    def get_yield_for_sample(
        self,
        region: Dict[str, Any],
        sample: Dict[str, Any],
        systematic: Optional[Dict[str, Any]] = None,
    ) -> List[float]:
        """Returns the yield for a specific sample.

        Args:
            region (Dict[str, Any]): specific region to use
            sample (Dict[str, Any]): specific sample to use
            systematic (Optional[Dict[str, Any]], optional): specific systematic
                variation to use, defaults to None -> {"Name": "Nominal"}

        Returns:
            List[float]: yields per bin for the sample
        """
        if systematic is None:
            systematic = {"Name": "Nominal"}

        histogram = histo.Histogram.from_config(
            self.histogram_folder, region, sample, systematic, modified=True
        )
        histo_yield = histogram.yields.tolist()
        return histo_yield

    def get_unc_for_sample(
        self,
        region: Dict[str, Any],
        sample: Dict[str, Any],
        systematic: Optional[Dict[str, Any]] = None,
    ) -> List[float]:
        """Returns the MC stat. uncertainty for a specific sample.

        Args:
            region (Dict[str, Any]): specific region to use
            sample (Dict[str, Any]): specific sample to use
            systematic (Optional[Dict[str, Any]], optional): specific systematic
                variation to use, defaults to None -> {"Name": "Nominal"}

        Returns:
            List[float]: statistical uncertainty of yield per bin for the sample
        """
        if systematic is None:
            systematic = {"Name": "Nominal"}

        histogram = histo.Histogram.from_config(
            self.histogram_folder, region, sample, systematic, modified=True
        )
        histo_stdev = histogram.stdev.tolist()
        return histo_stdev

    def get_NF_modifiers(self, sample: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Returns the list of NormFactor modifiers acting on a sample.

        Args:
            sample (Dict[str, Any]): specific sample to get NormFactor modifiers for

        Returns:
            List[Dict[str, Any]]: NormFactor modifiers for sample
        """
        modifiers = []
        for NormFactor in self.config["NormFactors"]:
            if configuration.sample_affected_by_modifier(sample, NormFactor):
                log.debug(
                    f"adding NormFactor {NormFactor['Name']} to sample {sample['Name']}"
                )
                modifiers.append(
                    {"data": None, "name": NormFactor["Name"], "type": "normfactor"}
                )
        return modifiers

    @staticmethod
    def get_Normalization_modifier(systematic: Dict[str, Any]) -> Dict[str, Any]:
        """Returns a normalization modifier (OverallSys in `HistFactory`).

        Args:
            systematic (Dict[str, Any]): systematic for which modifier is constructed

        Returns:
            Dict[str, Any]: single `normsys` modifier for ``pyhf`` workspace
        """
        modifier = {}
        modifier.update({"name": systematic["Name"]})
        modifier.update({"type": "normsys"})
        modifier.update(
            {
                "data": {
                    "hi": 1 + systematic["Up"]["Normalization"],
                    "lo": 1 + systematic["Down"]["Normalization"],
                }
            }
        )
        return modifier

    def get_NormPlusShape_modifiers(
        self,
        region: Dict[str, Any],
        sample: Dict[str, Any],
        systematic: Dict[str, Any],
    ) -> List[Dict[str, Any]]:
        """Returns modifiers for a correlated shape + normalization effect.

        For a variation including a correlated shape + normalization effect, this
        provides the `histosys` and `normsys` modifiers for ``pyhf`` (in `HistFactory`
        language, this corresponds to a `HistoSys` and an `OverallSys`). Symmetrization
        could happen either at this stage (this is the case currently), or somewhere
        earlier, such as during template postprocessing.

        Args:
            region (Dict[str, Any]): region the systematic variation acts in
            sample (Dict[str, Any]): sample the systematic variation acts on
            systematic (Dict[str, Any]): the systematic variation under consideration

        Returns:
            List[Dict[str, Any]]: a list with a ``pyhf`` `normsys` modifier and a
            `histosys` modifier
        """
        # load the systematic variation histogram
        histogram_up = histo.Histogram.from_config(
            self.histogram_folder,
            region,
            sample,
            systematic,
            modified=True,
            template="Up",
        )

        # also need the nominal histogram
        histogram_nominal = histo.Histogram.from_config(
            self.histogram_folder, region, sample, {"Name": "Nominal"}, modified=True
        )

        if systematic.get("Down", {}).get("Symmetrize", False):
            # add support for two-sided variations that do not require symmetrization
            # if symmetrization is desired, should support different implementations

            # symmetrization according to "method 1" from issue #26:
            # first normalization, then symmetrization

            # normalize the variation to the same yield as nominal
            norm_effect = histogram_up.normalize_to_yield(histogram_nominal)
            norm_effect_up = norm_effect
            norm_effect_down = 2 - norm_effect
            histo_yield_up = histogram_up.yields.tolist()
            log.debug(
                f"normalization impact of systematic {systematic['Name']} on sample "
                f"{sample['Name']} in region {region['Name']} is {norm_effect:.3f}"
            )
            # need another histogram that corresponds to the "down" variation, which is
            # 2*nominal - up
            histo_yield_down = (
                2 * histogram_nominal.yields - histogram_up.yields
            ).tolist()
        else:
            histogram_down = histo.Histogram.from_config(
                self.histogram_folder,
                region,
                sample,
                systematic,
                modified=True,
                template="Down",
            )
            norm_effect_up = sum(histogram_up.yields) / sum(histogram_nominal.yields)
            norm_effect_down = sum(histogram_down.yields) / sum(
                histogram_nominal.yields
            )
            # normalize templates to same yield as nominal
            histo_yield_up = list(histogram_up.yields / norm_effect_up)
            histo_yield_down = list(histogram_down.yields / norm_effect_down)

        # add the normsys
        modifiers = []
        norm_modifier = {}
        norm_modifier.update({"name": systematic["Name"]})
        norm_modifier.update({"type": "normsys"})
        norm_modifier.update({"data": {"hi": norm_effect_up, "lo": norm_effect_down}})
        modifiers.append(norm_modifier)

        # add the shape part in a histosys
        shape_modifier = {}
        shape_modifier.update({"name": systematic["Name"]})
        shape_modifier.update({"type": "histosys"})
        shape_modifier.update(
            {"data": {"hi_data": histo_yield_up, "lo_data": histo_yield_down}}
        )
        modifiers.append(shape_modifier)
        return modifiers

    def get_sys_modifiers(
        self, region: Dict[str, Any], sample: Dict[str, Any]
    ) -> List[Dict[str, Any]]:
        """Returns the list of all systematic modifiers acting on a sample.

        Args:
            region (Dict[str, Any]): region considered
            sample (Dict[str, Any]): specific sample to get modifiers for

        Raises:
            NotImplementedError: when unsupported modifiers act on sample

        Returns:
            List[Dict[str, Any]]: modifiers for ``pyhf`` workspace
        """
        modifiers = []
        for systematic in self.config.get("Systematics", []):
            if configuration.sample_affected_by_modifier(sample, systematic):
                if systematic["Type"] == "Normalization":
                    # OverallSys (norm uncertainty with Gaussian constraint)
                    log.debug(
                        f"adding OverallSys {systematic['Name']} to sample"
                        f" {sample['Name']}",
                    )
                    modifiers.append(self.get_Normalization_modifier(systematic))
                elif systematic["Type"] == "NormPlusShape":
                    # two modifiers are needed - an OverallSys for the norm effect,
                    # and a HistoSys for the shape variation
                    log.debug(
                        f"adding OverallSys and HistoSys {systematic['Name']} to sample"
                        f" {sample['Name']}",
                    )
                    modifiers += self.get_NormPlusShape_modifiers(
                        region, sample, systematic
                    )
                else:
                    raise NotImplementedError(
                        "not supporting other systematic types yet"
                    )
        return modifiers

    def get_channels(self) -> List[Dict[str, Any]]:
        """Returns the channel information: yields per sample and modifiers.

        Returns:
            List[Dict[str, Any]]: channels for ``pyhf`` workspace
        """
        channels = []
        for region in self.config["Regions"]:
            channel = {}
            channel.update({"name": region["Name"]})
            samples = []
            for sample in self.config["Samples"]:
                if sample.get("Data", False):
                    # skip the data sample, it goes into the observations instead
                    continue

                # yield of the samples
                histo_yield = self.get_yield_for_sample(region, sample)
                current_sample = {}
                current_sample.update({"name": sample["Name"]})
                current_sample.update({"data": histo_yield})

                # collect all modifiers for the sample
                modifiers = []

                # gammas
                stat_unc = self.get_unc_for_sample(region, sample)
                gammas = {}
                gammas.update({"name": "staterror_" + region["Name"].replace(" ", "-")})
                gammas.update({"type": "staterror"})
                gammas.update({"data": stat_unc})
                modifiers.append(gammas)

                # check if normfactor affect the sample and add modifiers as needed
                nf_modifier_list = self.get_NF_modifiers(sample)
                modifiers += nf_modifier_list

                # check if systematic uncertainties affect the samples, add modifiers
                sys_modifier_list = self.get_sys_modifiers(region, sample)
                modifiers += sys_modifier_list

                current_sample.update({"modifiers": modifiers})
                samples.append(current_sample)
            channel.update({"samples": samples})
            channels.append(channel)
        return channels

    def get_measurements(self) -> List[Dict[str, Any]]:
        """Returns the measurements object for the workspace.

        Constructs the measurements, including POI setting and parameter bounds, initial
        values and whether they are set to constant. Only supports a single measurement
        so far.

        Returns:
            List[Dict[str, Any]]: measurements for ``pyhf`` workspace
        """
        measurements = []
        measurement = {}
        measurement.update({"name": self.config["General"]["Measurement"]})
        config_dict = {}

        # get the norm factor initial values / bounds / constant setting
        parameters_list = []
        for nf in self.config.get("NormFactors", []):
            nf_name = nf["Name"]  # every NormFactor has a name
            init = nf.get("Nominal", None)
            bounds = nf.get("Bounds", None)
            fixed = self._get_constant_parameter_setting(nf_name)
            if (init is None) and (fixed is not None):
                # if no initial value is specified, but a constant setting
                # is requested, set the initial value to the constant value
                init = fixed

            parameter = {"name": nf_name}
            if init is not None:
                parameter.update({"inits": [init]})
            if bounds is not None:
                parameter.update({"bounds": [bounds]})
            if fixed is not None:
                parameter.update({"fixed": True})

            parameters_list.append(parameter)

        for sys in self.config.get("Systematics", []):
            # when there are many more systematics than NormFactors, it would be more
            # efficient to loop over fixed parameters and exclude all NormFactor related
            # ones to set all the remaining ones to constant (which are systematics)
            sys_name = sys["Name"]  # every systematic has a name
            fixed = self._get_constant_parameter_setting(sys_name)
            if fixed is not None:
                parameter = {"name": sys_name}
                parameter.update({"inits": [fixed]})
                parameter.update({"fixed": True})
                parameters_list.append(parameter)

        parameters = {"parameters": parameters_list}
        config_dict.update(parameters)
        config_dict.update({"poi": self.config["General"]["POI"]})
        measurement.update({"config": config_dict})
        measurements.append(measurement)
        return measurements

    def get_observations(self) -> List[Dict[str, Any]]:
        """Returns the observations (data yields) for the workspace.

        Returns:
            List[Dict[str, Any]]: observations for ``pyhf`` workspace
        """
        data_sample = self._get_data_sample()
        observations = []
        for region in self.config["Regions"]:
            observation = {}
            histo_yield = self.get_yield_for_sample(region, data_sample)
            observation.update({"name": region["Name"]})
            observation.update({"data": histo_yield})
            observations.append(observation)
        return observations

    def build(self) -> Dict[str, Any]:
        """Constructs a `HistFactory` workspace in ``pyhf`` format.

        Returns:
            Dict[str, Any]: ``pyhf``-compatible `HistFactory` workspace
        """
        ws: Dict[str, Any] = {}  # the workspace

        # channels
        channels = self.get_channels()
        ws.update({"channels": channels})

        # measurements
        measurements = self.get_measurements()
        ws.update({"measurements": measurements})

        # build observations
        observations = self.get_observations()
        ws.update({"observations": observations})

        # workspace version
        ws.update({"version": "1.0.0"})

        return ws


def build(config: Dict[str, Any], with_validation: bool = True) -> Dict[str, Any]:
    """Returns a `HistFactory` workspace in ``pyhf`` format.

    Args:
        config (Dict[str, Any]): cabinetry configuration
        with_validation (bool, optional): validate workspace validity with pyhf,
            defaults to True

    Returns:
        Dict[str, Any]: ``pyhf``-compatible `HistFactory` workspace
    """
    log.info("building workspace")

    ws_builder = WorkspaceBuilder(config)
    ws = ws_builder.build()

    if with_validation:
        validate(ws)
    return ws


def validate(ws: Dict[str, Any]) -> None:
    """Validates a workspace with ``pyhf``.

    Args:
        ws (Dict[str, Any]): the workspace to validate
    """
    pyhf.Workspace(ws)


def save(ws: Dict[str, Any], file_path_string: Union[str, pathlib.Path]) -> None:
    """Serializes a workspace to a file.

    Args:
        ws (Dict[str, Any]): ``pyhf``-compatible `HistFactory` workspace
        file_path_string (Union[str, pathlib.Path]): path to the file to save the
            workspace in
    """
    file_path = pathlib.Path(file_path_string)
    log.debug(f"saving workspace to {file_path}")
    # create output directory if it does not exist yet
    file_path.parent.mkdir(parents=True, exist_ok=True)
    # save workspace to file
    file_path.write_text(json.dumps(ws, sort_keys=True, indent=4))


def load(file_path_string: Union[str, pathlib.Path]) -> Dict[str, Any]:
    """Loads a workspace from a file.

    Args:
        file_path_string (Union[str, pathlib.Path]): path to the file to load the
            workspace from

    Returns:
        Dict[str, Any]: ``pyhf``-compatible `HistFactory` workspace
    """
    file_path = pathlib.Path(file_path_string)
    ws = json.loads(file_path.read_text())
    return ws



import fnmatch
import logging
from typing import Any, Callable, Dict, List, Optional, Union

import boost_histogram as bh

from . import configuration


log = logging.getLogger(__name__)


# type of a function processing templates, takes sample-region-systematic-template,
# returns None
ProcessorFunc = Callable[[Dict[str, Any], Dict[str, Any], Dict[str, Any], str], None]

# type of a user-defined function for template processing, takes sample-region-
# systematic-template, returns a boost_histogram.Histogram
UserTemplateFunc = Callable[
    [Dict[str, Any], Dict[str, Any], Dict[str, Any], str], bh.Histogram
]

# type of a generic function that takes sample-region-systematic-template, and
# may return None or a boost_histogram.Histogram
GenericProcessorFunc = Union[ProcessorFunc, UserTemplateFunc]

# type of a function called with names of region-sample-systematic-template,
# which returns either a ProcessorFunc or None
MatchFunc = Callable[[str, str, str, str], Optional[ProcessorFunc]]

# type of wrapper function that that turns a user-defined template processing function
# (which returns a histogram) into a function that returns None
WrapperFunc = Callable[[UserTemplateFunc], ProcessorFunc]


class Router:
    """Holds user-defined processing functions and matches functions to templates.

    Provides functions for matching a pattern to apply the right function to each
    template.

    Attributes:
        template_builders (List[Dict[str, Any]]): user-defined processors for template
            building
        template_builder_wrapper (Optional[WrapperFunc]): wrapper to apply on user-
            defined template builders
    """

    def __init__(self) -> None:
        """Initialize a Router instance, with no processors or wrappers defined."""
        # initialize all lists of processor types the user can specify
        self.template_builders: List[Dict[str, Any]] = []

        # initialize the wrapper to be used for template building to turn the user-
        # defined function (which returns a histogram) into one that saves the histogram
        # this wrapper needs to be set before using _find_template_builder_match to
        # properly handle the user-defined template builder
        self.template_builder_wrapper: Optional[WrapperFunc] = None

    @staticmethod
    def _register_processor(
        processor_list: List[Dict[str, Any]],
        region_name: Optional[str],
        sample_name: Optional[str],
        systematic_name: Optional[str],
        template: Optional[str],
    ) -> Callable[[GenericProcessorFunc], GenericProcessorFunc]:
        """Decorator for registering a custom processor function.

        Args:
            region_name  (Optional[str]): name of the region to apply the function to,
                or None to apply to all regions
            sample_name  (Optional[str]): name of the sample to apply the function to,
                or None to apply to all samples
            systematic_name  (Optional[str]): name of the systematic to apply the
                function to, or None to apply to all systematics
            template (Optional[str]): name of the template to apply the function to, or
                None to apply to all templates

        Returns:
            Callable[[GenericProcessorFunc], GenericProcessorFunc]: the function to
            register a processor
        """
        if region_name is None:
            region_name = "*"
        if sample_name is None:
            sample_name = "*"
        if systematic_name is None:
            systematic_name = "*"
        if template is None:
            template = "*"

        def _register(func: GenericProcessorFunc) -> GenericProcessorFunc:
            """Registers a processor function to be applied when matching a pattern.

            The pattern is specified by region-sample-systematic-template.

            Args:
                func (GenericProcessorFunc): the function to register
            """
            processor_list.append(
                {
                    "region": region_name,
                    "sample": sample_name,
                    "systematic": systematic_name,
                    "template": template,
                    "name": func.__name__,
                    "func": func,
                }
            )
            return func

        return _register

    def register_template_builder(
        self,
        region_name: Optional[str] = None,
        sample_name: Optional[str] = None,
        systematic_name: Optional[str] = None,
        template: Optional[str] = None,
    ) -> Callable[[Callable], UserTemplateFunc]:
        """Decorator for registering a template builder function.

        Args:
            region_name (Optional[str], optional): name of the region to apply the
                function to, defaults to None (apply to all regions)
            sample_name (Optional[str], optional): name of the sample to apply the
                function to, defaults to None (apply to all samples)
            systematic_name (Optional[str], optional): name of the systematic to apply
                the function to, defaults to None (apply to all systematics)
            template (Optional[str], optional): name of the template to apply the
                function to, defaults to None (apply to all templates)

        Returns:
            Callable[[Callable], UserTemplateFunc]: the generic function to register a
            processor
        """
        return self._register_processor(
            self.template_builders, region_name, sample_name, systematic_name, template
        )

    @staticmethod
    def _find_match(
        processor_list: List[Dict[str, Any]],
        region_name: str,
        sample_name: str,
        systematic_name: str,
        template: str,
    ) -> Optional[GenericProcessorFunc]:
        """Returns a function matching the provided specification.

        Args:
            processor_list (List[Dict[str, Any]]): list of processors to search in
            region_name (str): region name
            sample_name (str): sample name
            systematic_name (str): systematic name
            template (str): template name

        Returns:
            Optional[GenericProcessorFunc]: processor function matching the description,
            or None if no matches are found
        """
        matches = []
        for processor in processor_list:
            region_matches = fnmatch.fnmatch(region_name, processor["region"])
            sample_matches = fnmatch.fnmatch(sample_name, processor["sample"])
            systematic_matches = fnmatch.fnmatch(
                systematic_name, processor["systematic"]
            )
            template_matches = fnmatch.fnmatch(template, processor["template"])
            if (
                region_matches
                and sample_matches
                and systematic_matches
                and template_matches
            ):
                matches.append(processor["func"])

        if len(matches) > 1:
            log.warning(
                f"found {len(matches)} matches, continuing with the first one "
                f"({matches[0].__name__})"
            )
        elif len(matches) == 0:
            return None
        return matches[0]

    def _find_template_builder_match(
        self, region_name: str, sample_name: str, systematic_name: str, template: str
    ) -> Optional[ProcessorFunc]:
        """Returns a template builder function matching the provided specification.

        If no matches are found, returns None. Wraps the user-defined function in the
        provided wrapper, if no wrapper is found raises an error.

        Args:
            region_name (str): region name
            sample_name (str): sample name
            systematic_name (str): systematic name
            template (str): template name

        Raises:
            ValueError: when no template wrapper is set

        Returns:
            Optional[ProcessorFunc]: wrapped template builder function matching the
            description, or None if no matches are found
        """
        if self.template_builder_wrapper is None:
            # a wrapper needs to be defined already to convert the user-defined
            # UserTemplateFunc into a ProcessorFunc
            raise ValueError("no template builder wrapper defined")

        match = self._find_match(
            self.template_builders, region_name, sample_name, systematic_name, template
        )

        if match is not None:
            # if user-defined function was found, wrap and return it
            return self.template_builder_wrapper(match)
        return None


def apply_to_all_templates(
    config: Dict[str, Any],
    default_func: ProcessorFunc,
    match_func: Optional[MatchFunc] = None,
) -> None:
    """Applies the supplied function ``default_func`` to all templates.

    The templates are specified by the configuration file. The function takes four
    arguments in this order:

    - the dict specifying region information
    - the dict specifying sample information
    - the dict specifying systematic information
    - name of the template being considered: "Nominal", "Up", "Down"

    In addition it is possible to specify a function that returns custom overrides. If
    one is found for a given template, it is used instead of the default.

    Args:
        config (Dict[str, Any]): cabinetry configuration
        default_func (ProcessorFunc): function to be called for every template by
            default
        match_func: (Optional[MatchFunc], optional): function that returns user-defined
            functions to override the call to ``default_func``, defaults to None (then
            it is not used)
    """
    for region in config["Regions"]:
        log.debug(f"  in region {region['Name']}")

        for sample in config["Samples"]:
            log.debug(f"    reading sample {sample['Name']}")

            for systematic in [{"Name": "Nominal"}] + config["Systematics"]:

                # determine how many templates need to be considered
                if systematic["Name"] == "Nominal":
                    # only nominal template is needed
                    templates = ["Nominal"]
                else:
                    # systematics can have up and down template
                    templates = ["Up", "Down"]

                for template in templates:

                    # determine whether a histogram is needed for this
                    # specific combination of sample-region-systematic-template
                    histo_needed = configuration.histogram_is_needed(
                        region, sample, systematic, template
                    )

                    if not histo_needed:
                        # no further action is needed, continue with the next
                        # region-sample-systematic combination
                        continue

                    log.debug(
                        f"      variation {systematic['Name']}"
                        f"{' ' + template if template != 'Nominal' else ''}"
                    )

                    func_override = None
                    if match_func is not None:
                        # check whether a user-defined function was registered that
                        # matches this region-sample-systematic-template
                        func_override = match_func(
                            region["Name"], sample["Name"], systematic["Name"], template
                        )
                    if func_override is not None:
                        # call the user-defined function
                        log.debug(
                            f"executing user-defined override "
                            f"{func_override.__name__}"
                        )
                        func_override(region, sample, systematic, template)
                    else:
                        # call the provided default function
                        default_func(region, sample, systematic, template)



import logging
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Union

import iminuit
import numpy as np
import pyhf
import scipy.optimize
import scipy.stats

from . import model_utils


log = logging.getLogger(__name__)


class FitResults(NamedTuple):
    """Collects fit results in one object.

    Args:
        bestfit (numpy.ndarray): best-fit results of parameters
        uncertainty (numpy.ndarray): uncertainties of best-fit parameter results
        labels (List[str]): parameter labels
        corr_mat (np.ndarray): parameter correlation matrix
        best_twice_nll (float): -2 log(likelihood) at best-fit point
        goodess_of_fit (float, optional): goodness-of-fit p-value, defaults to -1
    """

    bestfit: np.ndarray
    uncertainty: np.ndarray
    labels: List[str]
    corr_mat: np.ndarray
    best_twice_nll: float
    goodness_of_fit: float = -1


class RankingResults(NamedTuple):
    """Collects nuisance parameter ranking results in one object.

    The best-fit results per parameter, the uncertainties, and the labels should not
    include the parameter of interest, since no impact for it is calculated.

    Args:
        bestfit (numpy.ndarray): best-fit results of parameters
        uncertainty (numpy.ndarray): uncertainties of best-fit parameter results
        labels (List[str]): parameter labels
        prefit_up (numpy.ndarray): pre-fit impact in "up" direction
        prefit_down (numpy.ndarray): pre-fit impact in "down" direction
        postfit_up (numpy.ndarray): post-fit impact in "up" direction
        postfit_down (numpy.ndarray): post-fit impact in "down" direction
    """

    bestfit: np.ndarray
    uncertainty: np.ndarray
    labels: List[str]
    prefit_up: np.ndarray
    prefit_down: np.ndarray
    postfit_up: np.ndarray
    postfit_down: np.ndarray


class ScanResults(NamedTuple):
    """Collects likelihood scan results in one object.

    Args:
        name (str): name of parameter in scan
        bestfit (float): best-fit parameter value from unconstrained fit
        uncertainty (float): uncertainty of parameter in unconstrained fit
        parameter_values (np.ndarray): parameter values used in scan
        delta_nlls (np.ndarray): -2 log(L) difference at each scanned point
    """

    name: str
    bestfit: float
    uncertainty: float
    parameter_values: np.ndarray
    delta_nlls: np.ndarray


class LimitResults(NamedTuple):
    """Collects parameter upper limit results in one object.

    Args:
        observed_limit (np.ndarray): observed limit
        expected_limit (np.ndarray): expected limit, including 1 and 2 sigma bands
        observed_CLs (np.ndarray): observed CLs values
        expected_CLs (np.ndarray): expected CLs values, including 1 and 2 sigma bands
        poi_values (np.ndarray): POI values used in scan
    """

    observed_limit: float
    expected_limit: np.ndarray
    observed_CLs: np.ndarray
    expected_CLs: np.ndarray
    poi_values: np.ndarray


def print_results(
    fit_results: FitResults,
) -> None:
    """Prints the best-fit parameter results and associated uncertainties.

    Args:
        fit_results (FitResults): results of fit to be printed
    """
    max_label_length = max([len(label) for label in fit_results.labels])
    log.info("fit results (with symmetric uncertainties):")
    for i, label in enumerate(fit_results.labels):
        log.info(
            f"{label.ljust(max_label_length)}: {fit_results.bestfit[i]: .4f} +/- "
            f"{fit_results.uncertainty[i]:.4f}"
        )


def _fit_model_pyhf(
    model: pyhf.pdf.Model,
    data: List[float],
    init_pars: Optional[List[float]] = None,
    fix_pars: Optional[List[bool]] = None,
    minos: Optional[List[str]] = None,
) -> FitResults:
    """Uses the ``pyhf.infer`` API to perform a maximum likelihood fit.

    Parameters set to be fixed in the model are held constant. The ``init_pars``
    argument allows to override the ``pyhf`` default initial parameter settings, and the
    ``fix_pars`` argument overrides which parameters are held constant.

    Args:
        model (pyhf.pdf.Model): the model to use in the fit
        data (List[float]): the data to fit the model to
        init_pars (Optional[List[float]], optional): list of initial parameter settings,
            defaults to None (use ``pyhf`` suggested inits)
        fix_pars (Optional[List[bool]], optional): list of booleans specifying which
            parameters are held constant, defaults to None (use ``pyhf`` suggestion)
        minos (Optional[List[str]], optional): runs the MINOS algorithm for all
            parameters specified in the list, defaults to None (does not run MINOS)

    Returns:
        FitResults: object storing relevant fit results
    """
    pyhf.set_backend("numpy", pyhf.optimize.minuit_optimizer(verbose=True))

    result, result_obj = pyhf.infer.mle.fit(
        data,
        model,
        init_pars=init_pars,
        fixed_params=fix_pars,
        return_uncertainties=True,
        return_result_obj=True,
    )

    bestfit = result[:, 0]
    uncertainty = result[:, 1]
    labels = model_utils.get_parameter_names(model)
    corr_mat = result_obj.minuit.np_matrix(correlation=True, skip_fixed=False)
    best_twice_nll = float(result_obj.fun)  # convert 0-dim np.ndarray to float

    fit_results = FitResults(bestfit, uncertainty, labels, corr_mat, best_twice_nll)

    if minos is not None:
        parameters_translated = []
        for minos_par in minos:
            par_index = model_utils._get_parameter_index(minos_par, labels)
            if par_index != -1:
                # pyhf does not hand over parameter names, all parameters are known as
                # x0, x1, etc.
                parameters_translated.append(f"x{par_index}")

        _run_minos(result_obj.minuit, parameters_translated, labels)

    return fit_results


def _fit_model_custom(
    model: pyhf.pdf.Model,
    data: List[float],
    init_pars: Optional[List[float]] = None,
    fix_pars: Optional[List[bool]] = None,
    minos: Optional[List[str]] = None,
) -> FitResults:
    """Uses ``iminuit`` directly to perform a maximum likelihood fit.

    Parameters set to be fixed in the model are held constant. The ``init_pars``
    argument allows to override the ``pyhf`` default initial parameter settings, and the
    ``fix_pars`` argument overrides which parameters are held constant.

    Args:
        model (pyhf.pdf.Model): the model to use in the fit
        data (List[float]): the data to fit the model to
        init_pars (Optional[List[float]], optional): list of initial parameter settings,
            defaults to None (use ``pyhf`` suggested inits)
        fix_pars (Optional[List[bool]], optional): list of booleans specifying which
            parameters are held constant, defaults to None (use ``pyhf`` suggestion)
        minos (Optional[List[str]], optional): runs the MINOS algorithm for all
            parameters specified in the list, defaults to None (does not run MINOS)

    Returns:
        FitResults: object storing relevant fit results
    """
    pyhf.set_backend("numpy", pyhf.optimize.minuit_optimizer(verbose=True))

    # use init_pars provided in function argument if they exist, else use default
    init_pars = init_pars or model.config.suggested_init()
    par_bounds = model.config.suggested_bounds()
    # use fix_pars provided in function argument if they exist, else use default
    fix_pars = fix_pars or model.config.suggested_fixed()

    labels = model_utils.get_parameter_names(model)

    # set initial step size to 0 for fixed parameters
    # this will cause the associated parameter uncertainties to be 0 post-fit
    step_size = [0.1 if not fix_pars[i_par] else 0.0 for i_par in range(len(init_pars))]

    def twice_nll_func(pars: np.ndarray) -> float:
        twice_nll = -2 * model.logpdf(pars, data)
        return twice_nll[0]

    m = iminuit.Minuit.from_array_func(
        twice_nll_func,
        init_pars,
        error=step_size,
        limit=par_bounds,
        fix=fix_pars,
        name=labels,
        errordef=1,
        print_level=1,
    )
    # decrease tolerance (goal: EDM < 0.002*tol*errordef), default tolerance is 0.1
    m.tol /= 10
    m.migrad()
    m.hesse()

    bestfit = m.np_values()
    uncertainty = m.np_errors()
    corr_mat = m.np_matrix(correlation=True, skip_fixed=False)
    best_twice_nll = m.fval

    fit_results = FitResults(bestfit, uncertainty, labels, corr_mat, best_twice_nll)

    if minos is not None:
        _run_minos(m, minos, labels)

    return fit_results


def _fit_model(
    model: pyhf.pdf.Model,
    data: List[float],
    init_pars: Optional[List[float]] = None,
    fix_pars: Optional[List[bool]] = None,
    minos: Optional[List[str]] = None,
    custom_fit: bool = False,
) -> FitResults:
    """Interface for maximum likelihood fits through ``pyhf.infer`` API or ``iminuit``.

    Parameters set to be fixed in the model are held constant. The ``init_pars``
    argument allows to override the ``pyhf`` default initial parameter settings, and the
    ``fix_pars`` argument overrides which parameters are held constant.

    Args:
        model (pyhf.pdf.Model): the model to use in the fit
        data (List[float]): the data to fit the model to
        init_pars (Optional[List[float]], optional): list of initial parameter settings,
            defaults to None (use ``pyhf`` suggested inits)
        fix_pars (Optional[List[bool]], optional): list of booleans specifying which
            parameters are held constant, defaults to None (use ``pyhf`` suggestion)
        minos (Optional[List[str]], optional): runs the MINOS algorithm for all
            parameters specified in the list, defaults to None (does not run MINOS)
        custom_fit (bool, optional): whether to use the ``pyhf.infer`` API or
            ``iminuit``, defaults to False (using ``pyhf.infer``)

    Returns:
        FitResults: object storing relevant fit results
    """
    if not custom_fit:
        # use pyhf infer API
        fit_results = _fit_model_pyhf(
            model, data, init_pars=init_pars, fix_pars=fix_pars, minos=minos
        )
    else:
        # use iminuit directly
        fit_results = _fit_model_custom(
            model, data, init_pars=init_pars, fix_pars=fix_pars, minos=minos
        )
    log.debug(f"-2 log(L) = {fit_results.best_twice_nll:.6f} at best-fit point")
    return fit_results


def _run_minos(minuit_obj: iminuit.Minuit, minos: List[str], labels: List[str]) -> None:
    """Determines parameter uncertainties for a list of parameters with MINOS.

    Args:
        minuit_obj (iminuit.Minuit): Minuit instance to use
        minos (List[str]): parameters for which MINOS is run
        labels (List[str]]): names of all parameters known to ``iminuit``, these names
            are used in output (may be the same as the names under which ``iminiuit``
            knows parameters)
    """
    for par_name in minos:
        # get index of current parameter in labels (to translate its name if iminuit
        # did not receive the parameter labels)
        par_index = model_utils._get_parameter_index(par_name, minuit_obj.parameters)
        if par_index == -1:
            # parameter not found, skip calculation
            continue
        log.info(f"running MINOS for {labels[par_index]}")
        minuit_obj.minos(var=par_name)

    log.info("MINOS results:")
    max_label_length = max([len(label) for label in labels])
    minos_unc = minuit_obj.np_merrors()
    for i_par, unc_down, unc_up in zip(range(len(labels)), minos_unc[0], minos_unc[1]):
        # the uncertainties are 0.0 by default if MINOS has not been run
        if unc_up != 0.0 or unc_down != 0.0:
            log.info(
                f"{labels[i_par].ljust(max_label_length)} = "
                f"{minuit_obj.np_values()[i_par]:.4f} -{unc_down:.4f} +{unc_up:.4f}"
            )


def _goodness_of_fit(
    model: pyhf.pdf.Model, data: List[float], best_twice_nll: float
) -> float:
    """Calculates goodness-of-fit p-value with a saturated model.

    Args:
        model (pyhf.pdf.Model): model used in the fit for which goodness-of-fit should
            be calculated
        data (List[float]): the observed data
        best_twice_nll (float): best-fit -2 log(likelihood) of fit for which goodness-
            of-fit should be calculated

    Returns:
        float: goodness-of-fit p-value
    """
    main_data, aux_data = model.fullpdf_tv.split(pyhf.tensorlib.astensor(data))
    # Poisson term: log Poisson(data|lambda=data), sum is over log likelihood of bins
    poisson_ll = sum(pyhf.tensorlib.poisson_dist(main_data).log_prob(main_data))
    # constraint term: log Gaussian(aux_data|parameters) etc.
    constraint_ll = model.constraint_logpdf(
        aux_data, pyhf.tensorlib.astensor(model.config.suggested_init())
    )
    saturated_nll = -(poisson_ll + constraint_ll)  # saturated likelihood

    log.info("calculating goodness-of-fit")
    delta_nll = best_twice_nll / 2 - saturated_nll
    log.debug(f"Delta NLL = {delta_nll:.6f}")

    # calculate difference in degrees of freedom between fits, given by the number
    # of bins minus the number of unconstrained parameters
    n_dof = sum(
        model.config.channel_nbins.values()
    ) - model_utils.unconstrained_parameter_count(model)
    log.debug(f"number of degrees of freedom: {n_dof}")
    p_val = scipy.stats.chi2.sf(2 * delta_nll, n_dof)
    log.info(f"p-value for goodness-of-fit test: {p_val*100:.2f}%")
    return p_val


def fit(
    spec: Dict[str, Any],
    asimov: bool = False,
    minos: Optional[Union[str, List[str]]] = None,
    goodness_of_fit: bool = False,
    custom_fit: bool = False,
) -> FitResults:
    """Performs a  maximum likelihood fit, reports and returns the results.

    The ``asimov`` flag allows to fit the Asimov dataset instead of observed data.
    Depending on the ``custom_fit`` keyword argument, this uses either the
    ``pyhf.infer`` API or ``iminuit`` directly for more control over the minimization.

    Args:
        spec (Dict[str, Any]): a ``pyhf`` workspace specification
        asimov (bool, optional): whether to fit the Asimov dataset, defaults to False
        minos (Optional[Union[str, List[str]]], optional): runs the MINOS algorithm for
            all parameters specified in the list, defaults to None (does not run MINOS)
        goodness_of_fit (bool, optional): calculate goodness of fit with a saturated
            model (perfectly fits data with shapefactors in all bins), defaults to False
        custom_fit (bool, optional): whether to use the ``pyhf.infer`` API or
            ``iminuit``, defaults to False (using ``pyhf.infer``)

    Returns:
        FitResults: object storing relevant fit results
    """
    log.info("performing maximum likelihood fit")

    model, data = model_utils.model_and_data(spec, asimov=asimov)

    # convert minos parameter to list, if a parameter is specified and is not a list
    if minos is not None and not isinstance(minos, list):
        minos = [minos]

    # perform fit
    fit_results = _fit_model(model, data, minos=minos, custom_fit=custom_fit)

    print_results(fit_results)

    if goodness_of_fit:
        # calculate goodness-of-fit with saturated model
        p_val = _goodness_of_fit(model, data, fit_results.best_twice_nll)
        fit_results = fit_results._replace(goodness_of_fit=p_val)

    return fit_results


def ranking(
    spec: Dict[str, Any],
    fit_results: FitResults,
    asimov: bool = False,
    custom_fit: bool = False,
) -> RankingResults:
    """Calculates the impact of nuisance parameters on the parameter of interest (POI).

    The impact is given by the difference in the POI between the nominal fit, and a fit
    where the nuisance parameter is held constant at its nominal value plus/minus its
    associated uncertainty. The ``asimov`` flag determines which dataset is used to
    calculate the impact. To calculate the proper Asimov impact, the ``fit_results``
    should come from a fit to the Asimov dataset. The "pre-fit impact" is obtained by
    varying the nuisance parameters by their uncertainty given by their constraint term.

    Args:
        spec (Dict[str, Any]): a ``pyhf`` workspace specification
        fit_results (FitResults): fit results to use for ranking
        asimov (bool, optional): whether to fit the Asimov dataset, defaults
            to False
        custom_fit (bool, optional): whether to use the ``pyhf.infer`` API or
            ``iminuit``, defaults to False (using ``pyhf.infer``)

    Returns:
        RankingResults: fit results for parameters, and pre- and post-fit impacts
    """
    model, data = model_utils.model_and_data(spec, asimov=asimov)
    labels = model_utils.get_parameter_names(model)
    prefit_unc = model_utils.get_prefit_uncertainties(model)
    nominal_poi = fit_results.bestfit[model.config.poi_index]

    # get default initial parameter settings / whether parameters are constant
    init_pars_default = model.config.suggested_init()
    fix_pars_default = model.config.suggested_fixed()

    all_impacts = []
    for i_par, label in enumerate(labels):
        if label == model.config.poi_name:
            continue  # do not calculate impact of POI on itself
        log.info(f"calculating impact of {label} on {labels[model.config.poi_index]}")

        # hold current parameter constant
        fix_pars = fix_pars_default.copy()
        fix_pars[i_par] = True

        parameter_impacts = []
        # calculate impacts: pre-fit up, pre-fit down, post-fit up, post-fit down
        for np_val in [
            fit_results.bestfit[i_par] + prefit_unc[i_par],
            fit_results.bestfit[i_par] - prefit_unc[i_par],
            fit_results.bestfit[i_par] + fit_results.uncertainty[i_par],
            fit_results.bestfit[i_par] - fit_results.uncertainty[i_par],
        ]:
            # can skip pre-fit calculation for unconstrained parameters (their
            # pre-fit uncertainty is set to 0), and pre- and post-fit calculation
            # for fixed parameters (both uncertainties set to 0 as well)
            if np_val == fit_results.bestfit[i_par]:
                log.debug(f"impact of {label} is zero, skipping fit")
                parameter_impacts.append(0.0)
            else:
                init_pars = init_pars_default.copy()
                init_pars[i_par] = np_val  # set value of current nuisance parameter
                fit_results_ranking = _fit_model(
                    model,
                    data,
                    init_pars=init_pars,
                    fix_pars=fix_pars,
                    custom_fit=custom_fit,
                )
                poi_val = fit_results_ranking.bestfit[model.config.poi_index]
                parameter_impact = poi_val - nominal_poi
                log.debug(
                    f"POI is {poi_val:.6f}, difference to nominal is "
                    f"{parameter_impact:.6f}"
                )
                parameter_impacts.append(parameter_impact)
        all_impacts.append(parameter_impacts)

    all_impacts_np = np.asarray(all_impacts)
    prefit_up = all_impacts_np[:, 0]
    prefit_down = all_impacts_np[:, 1]
    postfit_up = all_impacts_np[:, 2]
    postfit_down = all_impacts_np[:, 3]

    # remove parameter of interest from bestfit / uncertainty / labels
    # such that their entries match the entries of the impacts
    bestfit = np.delete(fit_results.bestfit, model.config.poi_index)
    uncertainty = np.delete(fit_results.uncertainty, model.config.poi_index)
    labels = np.delete(fit_results.labels, model.config.poi_index).tolist()

    ranking_results = RankingResults(
        bestfit, uncertainty, labels, prefit_up, prefit_down, postfit_up, postfit_down
    )
    return ranking_results


def scan(
    spec: Dict[str, Any],
    par_name: str,
    par_range: Optional[Tuple[float, float]] = None,
    n_steps: int = 11,
    asimov: bool = False,
    custom_fit: bool = False,
) -> ScanResults:
    """Performs a likelihood scan over the specified parameter.

    If no parameter range is specified, center the scan around the best-fit result for
    the parameter that is being scanned, and scan over twice its uncertainty in each
    direction. The reported likelihood values are the differences between -2 log(L) at
    each point in the scan and the global minimum.

    Args:
        spec (Dict[str, Any]): a ``pyhf`` workspace specification
        par_name (str): name of parameter to scan over
        par_range (Optional[Tuple[float, float]], optional): upper and lower bounds of
            parameter in scan, defaults to None (automatically determine bounds)
        n_steps (int, optional): number of steps in scan, defaults to 10
        asimov (bool, optional): whether to fit the Asimov dataset, defaults to False
        custom_fit (bool, optional): whether to use the ``pyhf.infer`` API or
            ``iminuit``, defaults to False (using ``pyhf.infer``)

    Raises:
        ValueError: if parameter is not found in model

    Returns:
        ScanResults: includes parameter name, scanned values and 2*log(likelihood)
        offset
    """
    model, data = model_utils.model_and_data(spec, asimov=asimov)
    labels = model_utils.get_parameter_names(model)
    init_pars = model.config.suggested_init()
    fix_pars = model.config.suggested_fixed()

    # get index of parameter with name par_name
    par_index = model_utils._get_parameter_index(par_name, labels)
    if par_index == -1:
        raise ValueError(f"could not find parameter {par_name} in model")

    # run a fit with the parameter not held constant, to find the best-fit point
    fit_results = _fit_model(model, data, custom_fit=custom_fit)
    nominal_twice_nll = fit_results.best_twice_nll
    par_mle = fit_results.bestfit[par_index]
    par_unc = fit_results.uncertainty[par_index]

    if par_range is None:
        # if no parameter range is specified, use +/-2 sigma from the MLE
        par_range = (par_mle - 2 * par_unc, par_mle + 2 * par_unc)

    scan_values = np.linspace(par_range[0], par_range[1], n_steps)
    delta_nlls = np.zeros_like(scan_values)  # holds results

    fix_pars[par_index] = True  # hold scan parameter constant in fits

    log.info(
        f"performing likelihood scan for {par_name} in range ({par_range[0]:.3f}, "
        f"{par_range[1]:.3f}) with {n_steps} steps"
    )
    for i_par, par_value in enumerate(scan_values):
        log.debug(f"performing fit with {par_name} = {par_value:.3f}")
        init_pars_scan = init_pars.copy()
        init_pars_scan[par_index] = par_value
        scan_fit_results = _fit_model(
            model,
            data,
            init_pars=init_pars_scan,
            fix_pars=fix_pars,
            custom_fit=custom_fit,
        )
        # subtract best-fit
        delta_nlls[i_par] = scan_fit_results.best_twice_nll - nominal_twice_nll

    scan_results = ScanResults(par_name, par_mle, par_unc, scan_values, delta_nlls)
    return scan_results


def limit(
    spec: Dict[str, Any],
    bracket: Optional[List[float]] = None,
    asimov: bool = False,
    tolerance: float = 0.01,
) -> LimitResults:
    """Calculates observed and expected 95% confidence level upper parameter limits.

    Limits are calculated for the parameter of interest (POI) defined in the workspace.
    Brent's algorithm is used to automatically determine POI values to be tested.

    Args:
        spec (Dict[str, Any]): a ``pyhf`` workspace specification
        bracket (Optional[List[float]], optional): the two POI values used to start the
            observed limit determination, example: ``[1.0, 2.0]`` (final POI values may
            lie outside this bracket), the two values must not be the same, defaults to
            None (then uses ``[0.5, 1.5]`` as default)
        asimov (bool, optional): whether to fit the Asimov dataset, defaults to False
        tolerance (float, optional): tolerance for convergence to CLs=0.05, defaults to
            0.01

    Raises:
        ValueError: if lower and upper bracket value are the same

    Returns:
        LimitResults: observed and expected limits, CLs values, and scanned points
    """
    if bracket is None:
        bracket = [0.5, 1.5]
    elif bracket[0] == bracket[1]:
        raise ValueError(f"the two bracket values must not be the same: " f"{bracket}")

    pyhf.set_backend("numpy", pyhf.optimize.minuit_optimizer(verbose=False))
    model, data = model_utils.model_and_data(spec, asimov=asimov)

    log.info(f"calculating upper limit for {model.config.poi_name}")

    # set lower POI bound to zero (for use with qmu_tilde)
    par_bounds = model.config.suggested_bounds()
    par_bounds[model.config.poi_index] = [0, par_bounds[model.config.poi_index][1]]
    log.debug("setting lower parameter bound for POI to 0")

    poi_list = []  # scanned POI values
    observed_CLs_list = []  # observed CLs values, one entry per scan point
    expected_CLs_list = []  # expected CLs values, 5 per point (with 1 and 2 sigma band)

    def _CLs_distance_to_crossing(
        poi: float,
        data: List[float],
        model: pyhf.pdf.Model,
        which_limit: int,
        limit_label: str,
    ) -> float:
        """Objective function to minimize in order to find CLs=0.05 crossing.

        Each observed and expected CLs result is also appended to lists, useful for
        visualization and to optimize the starting bracket for subsequent calculations.
        For POI values below 0, returns the maximum possible distance of 0.95.

        Args:
            poi (float): value for parameter of interest
            data (List[float]): data to fit to
            model (pyhf.pdf.Model): model to fit to data
            which_limit (int): which limit to run, 0: observed, 1: expected -2 sigma, 2:
                expected -1 sigma, 3: expected, 4: expected +1 sigma, 5: expected +2
                sigma
            limit_label (str): string to use when referring to the current limit

        Returns:
            float: absolute value of difference to CLs=0.05
        """
        if poi < 0:
            # no fit needed for negative POI value, return a default value
            log.debug(
                f"optimizer used {model.config.poi_name} = {poi:.4f}, skipping fit and "
                f"setting CLs = 1"
            )
            return 0.95  # corresponds to distance of CLs = 1 to target CLs = 0.05
        results = pyhf.infer.hypotest(
            poi,
            data,
            model,
            qtilde=True,
            return_expected_set=True,
            par_bounds=par_bounds,
        )
        observed = float(results[0])
        expected = np.asarray(results[1])
        poi_list.append(poi)
        observed_CLs_list.append(observed)
        expected_CLs_list.append(expected)
        current_CLs = np.hstack((observed, expected))[which_limit]
        log.debug(
            f"{model.config.poi_name} = {poi:.4f}, {limit_label} CLs = "
            f"{current_CLs:.4f}"
        )
        return np.abs(current_CLs - 0.05)

    # calculate all limits, one by one: observed, expected -2 sigma, expected -1 sigma,
    # expected, expected +1 sigma, expected +2 sigma
    limit_labels = [
        "observed",
        "expected -2 sigma",
        "expected -1 sigma",
        "expected",
        "expected +1 sigma",
        "expected +2 sigma",
    ]
    steps_total = 0
    all_limits = []
    all_converged = True
    for i_limit, limit_label in enumerate(limit_labels):
        log.info(f"determining {limit_label} upper limit")

        # find the 95% CL upper limit
        res = scipy.optimize.minimize_scalar(
            _CLs_distance_to_crossing,
            bracket=bracket,
            args=(data, model, i_limit, limit_label),
            method="brent",
            options={"xtol": tolerance, "maxiter": 100},
        )
        if (not res.success) or (res.fun > tolerance):
            log.error(
                f"failed to converge after {res.nfev} steps, distance from CLS=0.05 is "
                f"{res.fun:.4f}"
            )
            all_converged = False
        else:
            log.info(f"successfully converged after {res.nfev} steps")

        log.info(f"{limit_label} upper limit: {res.x:.4f}")
        all_limits.append(res.x)
        steps_total += res.nfev

        # determine the starting bracket for the next limit calculation
        if i_limit < 5:
            # get sorted list of POI values and associated expected CLs
            sorted_indices = np.argsort(poi_list)
            expected_CLs_np = np.asarray(expected_CLs_list)[sorted_indices]
            poi_list_np = np.asarray(poi_list)[sorted_indices]

            # interpolate to get expected CLs=0.06 and CLs=0.04 positions, inverted as
            # np.interp expects function to increase
            # for i_limit = 0, the next limit will be expected -2 sigma, corresponding
            # to expected_CLs_np[:, 0] etc.
            next_bracket: List[float] = np.interp(
                [0.06, 0.04], expected_CLs_np[:, i_limit][::-1], poi_list_np[::-1]
            ).tolist()
            # if the interpolation fails and the lower/upper bound are the same, then
            # offset both values to avoid getting stuck
            if next_bracket[0] == next_bracket[1]:
                next_bracket = [next_bracket[0] - 1, next_bracket[1] + 1]
            bracket = next_bracket

    # report all results
    log.info(f"total of {steps_total} steps to calculate all limits")
    if not all_converged:
        log.error("one or more calculations did not converge, check log")
    log.info("summary of upper limits:")
    for i_limit, limit_label in enumerate(limit_labels):
        log.info(f"{limit_label.ljust(17)}: {all_limits[i_limit]:.4f}")

    # sort all CLs values and scanned POI points by increasing POI value
    sorted_indices = np.argsort(poi_list)
    observed_CLs_np = np.asarray(observed_CLs_list)[sorted_indices]
    expected_CLs_np = np.asarray(expected_CLs_list)[sorted_indices]
    poi_list_np = np.asarray(poi_list)[sorted_indices]

    limit_results = LimitResults(
        all_limits[0],
        np.asarray(all_limits[1:]),
        observed_CLs_np,
        expected_CLs_np,
        poi_list_np,
    )
    return limit_results



import logging
import pathlib
from typing import Any, Dict, List, Optional, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np


log = logging.getLogger(__name__)


def data_MC(
    histogram_dict_list: List[Dict[str, Any]],
    total_model_unc: np.ndarray,
    bin_edges: np.ndarray,
    figure_path: pathlib.Path,
    log_scale: Optional[bool] = None,
) -> None:
    """Draws a data/MC histogram with uncertainty bands and ratio panel.

    Args:
        histogram_dict_list (List[Dict[str, Any]]): list of samples (with info stored in
            one dict per sample)
        total_model_unc (np.ndarray): total model uncertainty, if specified this is used
            instead of calculating it via sum in quadrature, defaults to None
        bin_edges (np.ndarray): bin edges of histogram
        figure_path (pathlib.Path): path where figure should be saved
        log_scale (Optional[bool], optional): whether to use a logarithmic vertical
            axis, defaults to None (automatically determine whether to use linear or log
            scale)
    """
    mc_histograms_yields = []
    mc_labels = []
    for h in histogram_dict_list:
        if h["isData"]:
            data_histogram_yields = h["yields"]
            data_histogram_stdev = np.sqrt(data_histogram_yields)
            data_label = h["label"]
        else:
            mc_histograms_yields.append(h["yields"])
            mc_labels.append(h["label"])

    mpl.style.use("seaborn-colorblind")

    fig = plt.figure(figsize=(6, 6))
    gs = fig.add_gridspec(nrows=2, ncols=1, hspace=0, height_ratios=[3, 1])
    ax1 = fig.add_subplot(gs[0])
    ax2 = fig.add_subplot(gs[1])

    # increase font sizes
    for item in (
        [ax1.yaxis.label, ax2.xaxis.label, ax2.yaxis.label]
        + ax1.get_yticklabels()
        + ax2.get_xticklabels()
        + ax2.get_yticklabels()
    ):
        item.set_fontsize("large")

    # minor ticks on all axes
    for axis in [ax1.xaxis, ax1.yaxis, ax2.xaxis, ax2.yaxis]:
        axis.set_minor_locator(mpl.ticker.AutoMinorLocator())

    # plot MC stacked together
    total_yield = np.zeros_like(mc_histograms_yields[0])
    bin_right_edges = bin_edges[1:]
    bin_left_edges = bin_edges[:-1]
    bin_width = bin_right_edges - bin_left_edges
    bin_centers = 0.5 * (bin_left_edges + bin_right_edges)
    mc_containers = []
    for mc_sample_yield in mc_histograms_yields:
        mc_container = ax1.bar(
            bin_centers,
            mc_sample_yield,
            width=bin_width,
            bottom=total_yield,
        )
        mc_containers.append(mc_container)

        # add a black line on top of each sample
        line_x = [y for y in bin_edges for _ in range(2)][1:-1]
        line_y = [y for y in (mc_sample_yield + total_yield) for _ in range(2)]
        ax1.plot(line_x, line_y, "-", color="black", linewidth=0.5)

        total_yield += mc_sample_yield

    # add total MC uncertainty
    mc_unc_container = ax1.bar(
        bin_centers,
        2 * total_model_unc,
        width=bin_width,
        bottom=total_yield - total_model_unc,
        fill=False,
        linewidth=0,
        edgecolor="gray",
        hatch=3 * "/",
    )

    # plot data
    data_container = ax1.errorbar(
        bin_centers,
        data_histogram_yields,
        yerr=data_histogram_stdev,
        fmt="o",
        color="k",
    )

    # ratio plot
    ax2.plot(
        [bin_left_edges[0], bin_right_edges[-1]],
        [1, 1],
        "--",
        color="black",
        linewidth=1,
    )  # reference line along y=1

    # add uncertainty band around y=1
    rel_mc_unc = total_model_unc / total_yield
    ax2.bar(
        bin_centers,
        2 * rel_mc_unc,
        width=bin_width,
        bottom=1 - rel_mc_unc,
        fill=False,
        linewidth=0,
        edgecolor="gray",
        hatch=3 * "/",
    )

    # data in ratio plot
    data_model_ratio = data_histogram_yields / total_yield
    data_model_ratio_unc = data_histogram_stdev / total_yield
    ax2.errorbar(
        bin_centers, data_model_ratio, yerr=data_model_ratio_unc, fmt="o", color="k"
    )

    # get the highest single bin yield, from the sum of MC or data
    y_max = max(
        np.max(total_yield),
        np.max([h["yields"] for h in histogram_dict_list if h["isData"]]),
    )
    # lowest MC yield in single bin (not considering empty bins)
    y_min = np.min(total_yield[np.nonzero(total_yield)])

    # determine scale setting, unless it is provided
    if log_scale is None:
        # if yields vary over more than 2 orders of magnitude, set y-axis to log scale
        log_scale = (y_max / y_min) > 100

    # set vertical axis scale and limits
    if log_scale:
        # use log scale
        ax1.set_yscale("log")
        ax1.set_ylim([y_min / 10, y_max * 10])
        # add "_log" to the figure name
        figure_path = figure_path.with_name(
            figure_path.stem + "_log" + figure_path.suffix
        )
    else:
        # do not use log scale
        ax1.set_ylim([0, y_max * 1.5])  # 50% headroom

    # MC contributions in inverse order, such that first legend entry corresponds to
    # the last (highest) contribution to the stack
    all_containers = mc_containers[::-1] + [mc_unc_container, data_container]
    all_labels = mc_labels[::-1] + ["Uncertainty", data_label]
    ax1.legend(all_containers, all_labels, frameon=False, fontsize="large")

    ax1.set_xlim(bin_left_edges[0], bin_right_edges[-1])
    ax1.set_ylabel("events")
    ax1.set_xticklabels([])
    ax1.tick_params(axis="both", which="major", pad=8)  # tick label - axis padding
    ax1.tick_params(direction="in", top=True, right=True, which="both")

    ax2.set_xlim(bin_left_edges[0], bin_right_edges[-1])
    ax2.set_ylim([0.5, 1.5])
    ax2.set_xlabel(histogram_dict_list[0]["variable"])
    ax2.set_ylabel("data / model")
    ax2.set_yticks([0.5, 0.75, 1.0, 1.25, 1.5])
    ax2.set_yticklabels([0.5, 0.75, 1.0, 1.25, ""])
    ax2.tick_params(axis="both", which="major", pad=8)
    ax2.tick_params(direction="in", top=True, right=True, which="both")

    fig.tight_layout()

    figure_path.parent.mkdir(parents=True, exist_ok=True)
    log.debug(f"saving figure as {figure_path}")
    fig.savefig(figure_path)
    plt.close(fig)


def correlation_matrix(
    corr_mat: np.ndarray,
    labels: Union[List[str], np.ndarray],
    figure_path: pathlib.Path,
) -> None:
    """Draws a correlation matrix.

    Args:
        corr_mat (np.ndarray): the correlation matrix to plot
        labels (Union[List[str], np.ndarray]): names of parameters in the correlation
            matrix
        figure_path (pathlib.Path): path where figure should be saved
    """
    # rounding for test in CI to match reference
    fig, ax = plt.subplots(
        figsize=(round(5 + len(labels) / 1.6, 1), round(3 + len(labels) / 1.6, 1)),
        dpi=100,
    )
    im = ax.imshow(corr_mat, vmin=-1, vmax=1, cmap="RdBu")

    ax.set_xticks(np.arange(len(labels)))
    ax.set_yticks(np.arange(len(labels)))
    ax.set_xticklabels(labels)
    ax.set_yticklabels(labels)
    for tick in ax.get_xticklabels():
        tick.set_rotation(45)
        tick.set_horizontalalignment("right")

    fig.colorbar(im, ax=ax)
    ax.set_aspect("auto")  # to get colorbar aligned with matrix
    fig.tight_layout()

    # add correlation as text
    for (j, i), corr in np.ndenumerate(corr_mat):
        text_color = "white" if abs(corr_mat[j, i]) > 0.75 else "black"
        if abs(corr) > 0.005:
            ax.text(i, j, f"{corr:.2f}", ha="center", va="center", color=text_color)

    figure_path.parent.mkdir(parents=True, exist_ok=True)
    log.debug(f"saving figure as {figure_path}")
    fig.savefig(figure_path)
    plt.close(fig)


def pulls(
    bestfit: np.ndarray,
    uncertainty: np.ndarray,
    labels: Union[List[str], np.ndarray],
    figure_path: pathlib.Path,
) -> None:
    """Draws a pull plot.

    Args:
        bestfit (np.ndarray): best-fit parameter results
        uncertainty (np.ndarray): parameter uncertainties
        labels (Union[List[str], np.ndarray]): parameter names
        figure_path (pathlib.Path): path where figure should be saved
    """
    num_pars = len(bestfit)
    y_positions = np.arange(num_pars)[::-1]
    fig, ax = plt.subplots(figsize=(6, 1 + num_pars / 4), dpi=100)
    ax.errorbar(bestfit, y_positions, xerr=uncertainty, fmt="o", color="black")

    ax.fill_between([-2, 2], -0.5, len(bestfit) - 0.5, color="yellow")
    ax.fill_between([-1, 1], -0.5, len(bestfit) - 0.5, color="limegreen")
    ax.vlines(0, -0.5, len(bestfit) - 0.5, linestyles="dotted", color="black")

    ax.set_xlim([-3, 3])
    ax.set_xlabel(r"$\left(\hat{\theta} - \theta_0\right) / \Delta \theta$")
    ax.set_ylim([-0.5, num_pars - 0.5])
    ax.set_yticks(y_positions)
    ax.set_yticklabels(labels)
    ax.xaxis.set_minor_locator(mpl.ticker.AutoMinorLocator())  # minor ticks
    ax.tick_params(axis="both", which="major", pad=8)
    ax.tick_params(direction="in", top=True, right=True, which="both")
    fig.tight_layout()

    figure_path.parent.mkdir(parents=True, exist_ok=True)
    log.debug(f"saving figure as {figure_path}")
    fig.savefig(figure_path)
    plt.close(fig)


def ranking(
    bestfit: np.ndarray,
    uncertainty: np.ndarray,
    labels: Union[List[str], np.ndarray],
    impact_prefit_up: np.ndarray,
    impact_prefit_down: np.ndarray,
    impact_postfit_up: np.ndarray,
    impact_postfit_down: np.ndarray,
    figure_path: pathlib.Path,
) -> None:
    """Draws a ranking plot.

    Args:
        bestfit (np.ndarray): best-fit parameter results
        uncertainty (np.ndarray): parameter uncertainties
        labels (Union[List[str], np.ndarray]): parameter labels
        impact_prefit_up (np.ndarray): pre-fit impact in "up" direction per parameter
        impact_prefit_down (np.ndarray): pre-fit impact in "down" direction per
            parameter
        impact_postfit_up (np.ndarray): post-fit impact in "up" direction per parameter
        impact_postfit_down (np.ndarray): post-fit impact in "down" direction per
            parameter
        figure_path (pathlib.Path): path where figure should be saved
    """
    num_pars = len(bestfit)

    mpl.style.use("seaborn-colorblind")
    fig, ax_pulls = plt.subplots(figsize=(8, 2.5 + num_pars * 0.45), dpi=100)
    ax_impact = ax_pulls.twiny()  # second x-axis with shared y-axis, used for pulls

    # since pull axis is below impact axis, flip them so pulls show up on top
    ax_pulls.set_zorder(1)  # pulls axis on top
    ax_pulls.patch.set_visible(False)  # pulls axis does not hide impact axis

    # increase font sizes
    for item in (
        [ax_pulls.yaxis.label, ax_pulls.xaxis.label, ax_impact.xaxis.label]
        + ax_pulls.get_yticklabels()
        + ax_pulls.get_xticklabels()
        + ax_impact.get_xticklabels()
    ):
        item.set_fontsize("large")

    # lines through pulls of -1, 0, 1 for orientation
    # line does not go all the way up to the top x-axis, since it
    # belongs to the bottom x-axis
    ax_pulls.vlines(
        -1, -1, num_pars - 0.5, linestyles="dashed", color="black", linewidth=0.75
    )
    ax_pulls.vlines(
        0, -1, num_pars - 0.5, linestyles="dashed", color="black", linewidth=0.75
    )
    ax_pulls.vlines(
        1, -1, num_pars - 0.5, linestyles="dashed", color="black", linewidth=0.75
    )

    y_pos = np.arange(num_pars)[::-1]

    # pre-fit up
    pre_up = ax_impact.barh(
        y_pos, impact_prefit_up, fill=False, linewidth=1, edgecolor="C0"
    )
    # pre-fit down
    pre_down = ax_impact.barh(
        y_pos, impact_prefit_down, fill=False, linewidth=1, edgecolor="C5"
    )
    # post-fit up
    post_up = ax_impact.barh(y_pos, impact_postfit_up, color="C0")
    # post-fit down
    post_down = ax_impact.barh(y_pos, impact_postfit_down, color="C5")
    # nuisance parameter pulls
    pulls = ax_pulls.errorbar(bestfit, y_pos, xerr=uncertainty, fmt="o", color="k")

    ax_pulls.set_xlabel(r"$\left(\hat{\theta} - \theta_0\right) / \Delta \theta$")
    ax_impact.set_xlabel(r"$\Delta \mu$")
    ax_pulls.set_xlim([-2, 2])
    ax_impact.set_xlim([-5, 5])
    ax_pulls.set_ylim([-1, num_pars])

    # impact axis limits: need largest pre-fit impact
    impact_max = np.max(np.abs(impact_prefit_up, impact_prefit_down))
    ax_impact.set_xlim([-impact_max * 1.1, impact_max * 1.1])

    # minor ticks
    for axis in [ax_pulls.xaxis, ax_impact.xaxis]:
        axis.set_minor_locator(mpl.ticker.AutoMinorLocator())

    ax_pulls.set_yticks(y_pos)
    ax_pulls.set_yticklabels(labels)

    ax_pulls.tick_params(direction="in", which="both")
    ax_impact.tick_params(direction="in", which="both")

    fig.legend(
        (pre_up, pre_down, post_up, post_down, pulls),
        (
            r"pre-fit impact: $\theta = \hat{\theta} + \Delta \theta$",
            r"pre-fit impact: $\theta = \hat{\theta} - \Delta \theta$",
            r"post-fit impact: $\theta = \hat{\theta} + \Delta \hat{\theta}$",
            r"post-fit impact: $\theta = \hat{\theta} - \Delta \hat{\theta}$",
            "pulls",
        ),
        frameon=False,
        loc="upper left",
        ncol=3,
        fontsize="large",
    )
    leg_space = 1.0 / (num_pars + 3) + 0.03
    fig.tight_layout(rect=[0, 0, 1.0, 1 - leg_space])  # make space for legend on top

    figure_path.parent.mkdir(parents=True, exist_ok=True)
    log.debug(f"saving figure as {figure_path}")
    fig.savefig(figure_path)
    plt.close(fig)


def templates(
    nominal_histo: Dict[str, np.ndarray],
    up_histo: Dict[str, np.ndarray],
    down_histo: Dict[str, np.ndarray],
    bin_edges: np.ndarray,
    variable: str,
    figure_path: pathlib.Path,
) -> None:
    """Draws a nominal template and the associated up/down variations.

    If a variation template is an empty dict, it is not drawn.

    Args:
        nominal_histo (Dict[str, np.ndarray]): the nominal template
        up_histo (Dict[str, np.ndarray]): the "up" variation
        down_histo (Dict[str, np.ndarray]): the "down" variation
        bin_edges (np.ndarray): bin edges of histogram
        variable (str): variable name for the horizontal axis
        figure_path (pathlib.Path): path where figure should be saved
    """
    bin_width = bin_edges[1:] - bin_edges[:-1]
    bin_centers = 0.5 * (bin_edges[:-1] + bin_edges[1:])

    mpl.style.use("seaborn-colorblind")
    fig = plt.figure()
    gs = fig.add_gridspec(nrows=2, ncols=1, hspace=0, height_ratios=[3, 1])
    ax1 = fig.add_subplot(gs[0])
    ax2 = fig.add_subplot(gs[1])

    # ratio plot line through unity and stat. uncertainty of nominal
    ax2.plot(
        [bin_edges[0], bin_edges[-1]],
        [1, 1],
        "--",
        color="black",
        linewidth=1,
    )
    rel_nominal_stat_unc = nominal_histo["stdev"] / nominal_histo["yields"]
    ax2.bar(
        bin_centers,
        2 * rel_nominal_stat_unc,
        width=bin_width,
        bottom=1 - rel_nominal_stat_unc,
        fill=False,
        linewidth=0,
        edgecolor="gray",
        hatch=3 * "/",
    )

    colors = ["black", "C0", "C1"]
    linestyles = ["-", "--", "-."]
    all_templates = [nominal_histo, up_histo, down_histo]
    labels = ["nominal", "up", "down"]

    # x positions for lines drawn showing the template distributions
    line_x = [y for y in bin_edges for _ in range(2)][1:-1]

    # draw templates
    for template, color, linestyle, label in zip(
        all_templates, colors, linestyles, labels
    ):
        if not template:
            # variation not defined
            continue

        # lines to show each template distribution
        line_y = [y for y in template["yields"] for _ in range(2)]

        ax1.plot(
            line_x,
            line_y,
            "-",
            color=color,
            linestyle=linestyle,
            label=label,
        )
        if label == "nominal":
            # band for stat. uncertainty of nominal prediction
            ax1.bar(
                bin_centers,
                2 * nominal_histo["stdev"],
                width=bin_width,
                bottom=nominal_histo["yields"] - nominal_histo["stdev"],
                fill=False,
                linewidth=0,
                edgecolor="gray",
                hatch=3 * "/",
            )
        else:
            # error bars for up/down variations
            ax1.errorbar(
                bin_centers,
                template["yields"],
                yerr=template["stdev"],
                fmt="none",
                color=color,
            )

            # ratio plot: variation / nominal
            template_ratio_plot = template["yields"] / nominal_histo["yields"]
            line_y = [y for y in template_ratio_plot for _ in range(2)]

            ax2.plot(
                line_x,
                line_y,
                "-",
                color=color,
                linestyle=linestyle,
            )
            ax2.errorbar(
                bin_centers,
                template_ratio_plot,
                yerr=template["stdev"] / nominal_histo["yields"],
                fmt="none",
                color=color,
            )

    # increase font sizes
    for item in (
        [ax1.yaxis.label, ax2.xaxis.label, ax2.yaxis.label]
        + ax1.get_yticklabels()
        + ax2.get_xticklabels()
        + ax2.get_yticklabels()
    ):
        item.set_fontsize("large")

    # minor ticks on all axes
    for axis in [ax1.xaxis, ax1.yaxis, ax2.xaxis, ax2.yaxis]:
        axis.set_minor_locator(mpl.ticker.AutoMinorLocator())

    ax1.legend(frameon=False, fontsize="large")

    max_yield = max([max(template["yields"]) for template in all_templates if template])

    ax1.set_xlim([bin_edges[0], bin_edges[-1]])
    ax1.set_ylim([0, max_yield * 1.5])
    ax1.set_ylabel("events")
    ax1.set_xticklabels([])
    ax1.tick_params(axis="both", which="major", pad=8)  # tick label - axis padding
    ax1.tick_params(direction="in", top=True, right=True, which="both")

    ax2.set_xlim([bin_edges[0], bin_edges[-1]])
    ax2.set_ylim([0.5, 1.5])
    ax2.set_xlabel(variable)
    ax2.set_ylabel("variation / nominal")
    ax2.set_yticks([0.5, 0.75, 1.0, 1.25, 1.5])
    ax2.set_yticklabels([0.5, 0.75, 1.0, 1.25, ""])
    ax2.tick_params(axis="both", which="major", pad=8)
    ax2.tick_params(direction="in", top=True, right=True, which="both")

    fig.tight_layout()

    figure_path.parent.mkdir(parents=True, exist_ok=True)
    log.debug(f"saving figure as {figure_path}")
    fig.savefig(figure_path)
    plt.close(fig)


def scan(
    par_name: str,
    par_mle: float,
    par_unc: float,
    par_vals: np.ndarray,
    par_nlls: np.ndarray,
    figure_path: pathlib.Path,
) -> None:
    """Draws a figure showing the results of a likelihood scan.

    Args:
        par_name (str): name of parameter used in scan
        par_mle (float): best-fit result for parameter
        par_unc (float): best-fit parameter uncertainty
        par_vals (np.ndarray): values used in scan over parameter
        par_nlls (np.ndarray): -2 log(L) offset at each scan point
        figure_path (pathlib.Path): path where figure should be saved
    """
    mpl.style.use("seaborn-colorblind")
    fig, ax = plt.subplots()

    # line through y=1 and y=4 to show confidence intervals
    ax.plot([par_vals[0], par_vals[-1]], [1, 1], ":", color="gray")
    ax.plot([par_vals[0], par_vals[-1]], [4, 4], ":", color="gray")

    # position for text - right edge of the figure, with slight padding
    text_x_pos = par_vals[-1] - 0.01 * (par_vals[-1] - par_vals[0])
    ax.text(
        text_x_pos,
        1.0,
        "68% CL",
        horizontalalignment="right",
        verticalalignment="bottom",
        color="gray",
    )
    ax.text(
        text_x_pos,
        4.0,
        "95% CL",
        horizontalalignment="right",
        verticalalignment="bottom",
        color="gray",
    )

    # Gaussian at best-fit parameter value for reference
    val_grid = np.linspace(par_vals[0], par_vals[-1], 100)
    gaussian_approx = [((par_val - par_mle) / par_unc) ** 2 for par_val in val_grid]
    ax.plot(val_grid, gaussian_approx, "--", color="C5", label="Gaussian approximation")

    # scan results
    ax.plot(par_vals, par_nlls, "-", color="C0")
    ax.plot(par_vals, par_nlls, "X", color="C0", label="parameter scan")

    # increase font sizes
    for item in (
        [ax.xaxis.label, ax.yaxis.label] + ax.get_xticklabels() + ax.get_yticklabels()
    ):
        item.set_fontsize("large")

    # minor ticks
    for axis in [ax.xaxis, ax.yaxis]:
        axis.set_minor_locator(mpl.ticker.AutoMinorLocator())

    ax.set_xlabel(par_name)
    ax.set_xlim(par_vals[0], par_vals[-1])
    ax.set_ylabel(r"$-2 \Delta \log(L)$")
    ax.set_ylim(0, max(par_nlls) * 1.2)
    ax.tick_params(axis="both", which="major", pad=8)
    ax.tick_params(direction="in", top=True, right=True, which="both")

    ax.legend(frameon=False, fontsize="large")

    fig.tight_layout()

    figure_path.parent.mkdir(parents=True, exist_ok=True)
    log.debug(f"saving figure as {figure_path}")
    fig.savefig(figure_path)
    plt.close(fig)


def limit(
    observed_CLs: np.ndarray,
    expected_CLs: np.ndarray,
    poi_values: np.ndarray,
    figure_path: pathlib.Path,
) -> None:
    """Draws observed and expected CLs values as function of the parameter of interest.

    Args:
        observed_CLs (np.ndarray): observed CLs values
        expected_CLs (np.ndarray): expected CLs values, including 1 and 2 sigma bands
        poi_values (np.ndarray): parameter of interest values used in scan
        figure_path (pathlib.Path): path where figure should be saved
    """
    fig, ax = plt.subplots()

    xmin = min(poi_values)
    xmax = max(poi_values)

    # line through CLs = 0.05
    ax.hlines(
        0.05,
        xmin=xmin,
        xmax=xmax,
        linestyle="dashdot",
        color="red",
        label=r"CL$_S$ = 5%",
    )

    # 1 and 2 sigma bands
    ax.fill_between(
        poi_values,
        expected_CLs[:, 0],
        expected_CLs[:, 4],
        color="yellow",
        label=r"expected CL$_S$ $\pm 2\sigma$",
    )
    ax.fill_between(
        poi_values,
        expected_CLs[:, 1],
        expected_CLs[:, 3],
        color="limegreen",
        label=r"expected CL$_S$ $\pm 1\sigma$",
    )

    # expected CLs
    ax.plot(
        poi_values,
        expected_CLs[:, 2],
        "--",
        color="black",
        label=r"expected CL$_S$",
    )

    # observed CLs values
    ax.plot(poi_values, observed_CLs, "o-", color="black", label=r"observed CL$_S$")

    # increase font sizes
    for item in (
        [ax.xaxis.label, ax.yaxis.label] + ax.get_xticklabels() + ax.get_yticklabels()
    ):
        item.set_fontsize("large")

    # minor ticks
    for axis in [ax.xaxis, ax.yaxis]:
        axis.set_minor_locator(mpl.ticker.AutoMinorLocator())

    ax.legend(frameon=False, fontsize="large")

    ax.set_xlabel(r"$\mu$")
    ax.set_ylabel(r"$\mathrm{CL}_{s}$")
    ax.set_xlim([xmin, xmax])
    ax.set_ylim([0, 1])
    ax.tick_params(axis="both", which="major", pad=8)
    ax.tick_params(direction="in", top=True, right=True, which="both")

    fig.tight_layout()

    figure_path.parent.mkdir(parents=True, exist_ok=True)
    log.debug(f"saving figure as {figure_path}")
    fig.savefig(figure_path)
    plt.close(fig)
